{ File   : DeLZSS.dpr
  Author : Deniz Oezmen
  Created: 2006-10-09
  Changed: 2006-10-10

  Pretty stupid and inefficient LZSS decompressor (N = 12, F = 4).
}
program DeLZSS;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TBuffer = array of Byte;

function OpenInFile(var InFile: file; Name: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, Name);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  OpenInFile := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; Name: string): Boolean;
begin
  FileMode := fmOpenWrite;
  AssignFile(OutFile, Name);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  OpenOutFile := (IOResult = 0)
end;

procedure DecompressBlock(const InBuf: TBuffer; out OutBuf: TBuffer);
const
  NBits     = 12;
  FBits     = 4;
  Threshold = 3;
var
  N,
  F,
  Offset,
  Len      : Word;
  Flags,
  b        : Byte;
  i, j,
  InBufPos,
  OutBufPos,
  HisBufPos: Longint;
  HisBuf   : TBuffer;
begin
  // initialize LZSS parameters for this block
  N := 1 shl NBits - 1;
  F := 1 shl FBits - 1;
  HisBufPos := N - F - 2;

  SetLength(HisBuf, N + 1);
  FillChar(HisBuf[0], N + 1, 0);

  SetLength(OutBuf, 0);
  OutBufPos := 0;

  InBufPos := 0;

  // start extracting block contents
  while InBufPos < Length(InBuf) do
  begin
    // read bit field
    Flags := InBuf[InBufPos];
    Inc(InBufPos);

    // process bit field
    for i := 1 to 8 do
    begin
      if InBufPos < Length(InBuf) then
      begin
        // check for buffer window reference
        if (Flags and 1 = 0) then
        begin
          // read offset and length
          Len := InBuf[InBufPos + 1] and F + Threshold;
          Offset := ((InBuf[InBufPos + 1] shr FBits) shl 8 + InBuf[InBufPos]);

          Inc(InBufPos, 2);

          if (Offset = 0) and (Len = 0) then
            Exit;
            
          // peek into buffer
          SetLength(OutBuf, Length(OutBuf) + Len);
          for j := 0 to Len - 1 do
          begin
            b := HisBuf[(Offset + j) and N];
            OutBuf[OutBufPos] := b;
            Inc(OutBufPos);
            HisBuf[HisBufPos] := b;
            HisBufPos := (HisBufPos + 1) and N
          end
        end
        else
        // copy literally
        begin
          SetLength(OutBuf, Length(OutBuf) + 1);
          b := InBuf[InBufPos];
          Inc(InBufPos);
          OutBuf[OutBufPos] := b;
          HisBuf[HisBufPos] := b;
          Inc(OutBufPos);
          HisBufPos := (HisBufPos + 1) and N
        end;

        Flags := Flags shr 1
      end
    end
  end
end;

procedure Decompress(var InFile, OutFile: file);
var
  InBuf,
  OutBuf: TBuffer;
begin
  SetLength(InBuf, FileSize(InFile));

  BlockRead(InFile, InBuf[0], Length(InBuf));
  DecompressBlock(InBuf, OutBuf);
  BlockWrite(OutFile, OutBuf[0], Length(OutBuf))
end;

procedure ParseFiles(FilePara, OutDir: string);
var
  DirPos,
  i        : Byte;
  Dir      : string;
  SearchRec: TSearchRec;
  InFile,
  OutFile  : file;
begin
  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // check output directory
  if OutDir[Length(OutDir)] <> '\' then
    OutDir := OutDir + '\';
  if not DirectoryExists(OutDir) then
  begin
    WriteLn('Error: Output directory not accessible.');
    Exit
  end;

  // search for all files matching FilePara
  if FindFirst(FilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  repeat
    WriteLn(SearchRec.Name);

    if not OpenInFile(InFile, Dir + SearchRec.Name) then
    begin
      WriteLn('  not accessible');
      Continue
    end;

    if not OpenOutFile(OutFile, OutDir + SearchRec.Name) then
    begin
      WriteLn('  write error');
      Close(InFile);
      Continue
    end;

    Decompress(InFile, OutFile);

    Close(InFile);
    Close(OutFile)
  until FindNext(SearchRec) <> 0
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount <> 2 then
  begin
    Writeln(Format('Usage: %s <files> <Output>',
      [ExtractFileName(ParamStr(0))]));
    Exit
  end;

  ParseFiles(ParamStr(1), ParamStr(2))
end.
